import json
import os
from typing import List, Optional
import uuid
from tonic_validate.classes.user_info import UserInfo
from tonic_validate.utils.http_client import HttpClient
from appdirs import user_data_dir
from tonic_ragas_logger.config import Config
APP_DIR_NAME = "tonic-ragas-logger"

# List of CI/CD environment variables
# 1. Github Actions: GITHUB_ACTIONS
# 2  Gitlab CI/CD: GITLAB_CI
# 3. Azure devops: TF_BUILD
# 4. CircleCI: CI
# 5. Jenkins: JENKINS_URL
# 6. TravisCI: CI
# 7. Bitbucket: CI
env_vars = ["GITHUB_ACTIONS", "GITLAB_CI", "TF_BUILD", "CI", "JENKINS_URL"]


class Telemetry:
    def __init__(self, api_key: Optional[str] = None):
        """
        Used to log telemetry data to the Tonic Validate server

        Parameters
        ----------
        api_key: Optional[str]
            The API key to use for authentication
        """
        self.config = Config()
        self.http_client = HttpClient(self.config.TONIC_VALIDATE_TELEMETRY_URL, api_key)

    def get_user(self) -> UserInfo:
        """
        Retrieves the user information from the file. If the user does not exist, creates a new user

        Returns
        -------
        UserInfo
            Information about the user
        """
        app_dir_path = user_data_dir(appname=APP_DIR_NAME)
        user_id_path = os.path.join(app_dir_path, "user.json")
        # check if user_id exists else we create a new uuid and write it to the file
        if os.path.exists(user_id_path):
            with open(user_id_path, "r") as f:
                user_info = json.load(f)
        else:
            user_info: UserInfo = {"user_id": str(uuid.uuid4()), "linked": False}
            json_info = json.dumps(user_info)
            # create the directory if it does not exist
            os.makedirs(app_dir_path, exist_ok=True)
            with open(user_id_path, "w") as f:
                f.write(json_info)
        return user_info

    def __is_ci(self):
        """
        Checks whether the current environment is a CI/CD environment
        """
        for var in env_vars:
            if os.environ.get(var):
                return True
        return False

    def log_run(self, num_of_questions: int, metrics: List[str]):
        """
        Logs a run to the Tonic Validate server

        Parameters
        ----------
        num_of_questions: int
            The number of questions asked
        metrics: List[str]
            The metrics that were used to evaluate the run
        """
        if self.config.TONIC_RAGAS_DO_NOT_TRACK:
            return

        try:
            from importlib.metadata import version
            sdk_version = version('tonic-ragas-logger')
        except Exception:
            sdk_version = "unknown"

        user_id = self.get_user()["user_id"]
        self.http_client.http_post(
            "/runs",
            data={
                "user_id": user_id,
                "num_of_questions": num_of_questions,
                "metrics": metrics,
                "is_ci": self.__is_ci(),
                "backend": "ragas",
                "run_time": -1,
                "sdk_version": sdk_version
            },
            timeout=5,
        )


import os
from dotenv import load_dotenv


class Config:
    def __init__(self) -> None:
        load_dotenv()

        self.TONIC_VALIDATE_API_KEY = os.getenv(
            "TONIC_VALIDATE_API_KEY",
        )
        self.TONIC_VALIDATE_BASE_URL = os.getenv(
            "TONIC_VALIDATE_BASE_URL", "https://validate.tonic.ai/api/v1"
        )
        self.TONIC_VALIDATE_TELEMETRY_URL = os.getenv(
            "TONIC_VALIDATE_TELEMETRY_URL", "https://telemetry.tonic.ai/validate"
        )
        self.TONIC_RAGAS_DO_NOT_TRACK = os.getenv(
            "TONIC_RAGAS_DO_NOT_TRACK", "false"
        ).lower() in ("true", "1", "t")


from typing import Any, List, Optional, Dict

from tonic_ragas_logger.utils.telemetry import Telemetry
from tonic_validate import Run, RunData
from tonic_ragas_logger.config import Config

from tonic_ragas_logger.utils.http_client import HttpClient
from ragas.evaluation import Result


class RagasValidateApi:
    """Wrapper class for invoking the Tonic Validate UI.

    Parameters
    ----------
    api_key : str
        The access token for the Tonic Validate UI. The access token is obtained from
        the Tonic Validate UI.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
    ):
        self.config = Config()
        if api_key is None:
            api_key = self.config.TONIC_VALIDATE_API_KEY
            if api_key is None:
                exception_message = (
                    "No api key provided. Please provide an api key or set "
                    "TONIC_VALIDATE_API_KEY environment variable."
                )
                raise Exception(exception_message)
        self.client = HttpClient(self.config.TONIC_VALIDATE_BASE_URL, api_key)
        self.telemetry = Telemetry()

    def upload_results(
        self,
        project_id: str,
        results: Result,
        run_metadata: Dict[str, Any] = {},
        tags: List[str] = [],
    ) -> str:
        """Uploads results to a Tonic Validate project.

        Parameters
        ----------
        project_id : str
            The ID of the project to upload the run to.
        results : Result
            The result to upload.
        run_metadata : Dict[str, str]
            Metadata to attach to the run. If the values are not strings, then they are
            converted to strings before making the request.
        """
        run = self.__convert_to_run(results)
        run_response = self.client.http_post(
            f"/projects/{project_id}/runs/with_data",
            data={
                "run_metadata": run_metadata,
                "tags": tags,
                "data": [run_data.to_dict() for run_data in run.run_data],
            },
        )

        try:
            self.telemetry.log_run(
                len(run.run_data), list(run.overall_scores.keys())
            )
        except Exception as _:
            pass

        return run_response["id"]

    def __convert_to_run(self, results: Result) -> Run:
        """Converts a Result to a Run.

        Parameters
        ----------
        results : Result
            The result to convert.
        """
        try:
            overall_scores = {
                str(score): 0 if value is None else float(value)
                for score, value in results.items()
            }
        except ValueError:
            raise ValueError(
                "The scores in the results are in the format of {score: value} where score is a string and value is a float."
            )

        if not results.dataset:
            raise ValueError(
                "The ragas results do not have a dataset provided. Can not upload results without a dataset."
            )

        if len(results.scores) != len(results.dataset):
            raise ValueError(
                "The length of results.scores and results.dataset are not the same"
            )

        run_data = []
        for i in range(len(results.scores)):
            try:
                scores: Dict[str, float | None] = {
                    str(score): 0 if value is None else float(value)
                    for score, value in results.scores[i].items()
                }
            except ValueError:
                raise ValueError(
                    "The scores in the results are in the format of {score: value} where score is a string and value is a float."
                )
            run_data.append(
                RunData(
                    scores=scores,
                    reference_question=results.dataset["question"][i],
                    reference_answer=results.dataset["ground_truth"][i],
                    llm_answer=results.dataset["answer"][i],
                    llm_context=results.dataset["contexts"][i],
                )
            )

        return Run(
            overall_scores=overall_scores,
            run_data=run_data,
            id=None,
        )



from typing import Any, Dict, Optional, Union
import requests
from urllib3.exceptions import InsecureRequestWarning  # type: ignore

requests.packages.urllib3.disable_warnings(  # type: ignore
    category=InsecureRequestWarning
)


class HttpClient:
    """Client for handling requests to Tonic Validate instance.

    Parameters
    ----------
    base_url : str
        URL to Tonic Validate instance.
    access_token : str
        The API token associated with your Tonic Validate account.
    """

    def __init__(self, base_url: str, access_token: Optional[str] = None):
        self.base_url = base_url
        self.headers = None
        if access_token is not None:
            self.headers = {"Authorization": f"Bearer {access_token}"}

    def http_get(
        self, url: str, params: Dict[Any, Any] = {}, timeout: Union[int, None] = None
    ) -> Any:
        """Make a get request.

        Parameters
        ----------
        url : str
            URL to make get request. Is appended to self.base_url.
        params: dict
            Passed as the params parameter of the requests.get request.

        """
        res = requests.get(
            self.base_url + url,
            params=params,
            headers=self.headers,
            verify=False,
            timeout=timeout,
        )
        res.raise_for_status()
        return res.json()

    def http_post(
        self,
        url: str,
        params: Dict[Any, Any] = {},
        data: Dict[Any, Any] = {},
        timeout: Union[int, None] = None,
    ) -> Any:
        """Make a post request.

        Parameters
        ----------
        url : str
            URL to make the post request. Is appended to self.base_url.
        params: dict
            Passed as the params parameter of the requests.post request.
        data: dict
            Passed as the data parameter of the requests.post request.
        """
        res = requests.post(
            self.base_url + url,
            params=params,
            json=data,
            headers=self.headers,
            verify=False,
            timeout=timeout,
        )
        res.raise_for_status()
        return res.json()

    def http_put(
        self, url: str, params: Dict[Any, Any] = {}, data: Dict[Any, Any] = {}
    ) -> Any:
        """Make a put request.

        Parameters
        ----------
        url : str
            URL to make the put request. Is appended to self.base_url.
        params: dict
            Passed as the params parameter of the requests.put request.
        data: dict
            Passed as the data parameter of the requests.put request.
        """
        res = requests.put(
            self.base_url + url,
            params=params,
            json=data,
            headers=self.headers,
            verify=False,
        )
        res.raise_for_status()
        return res.json()
